from openai import OpenAI

def main():
    client = OpenAI(base_url="http://127.0.0.1:8000", api_key="secret_key")
    
    # Create new thread
    thread = client.beta.threads.create()

    while True:
        user_input = input("User: ")

        if user_input.lower() in ["quit", "exit", "q"]:
            print("Goodbye!")
            break

        # Add user message to thread
        client.beta.threads.messages.create(thread.id, role="user", content=user_input)

        # Run the thread
        client.beta.threads.runs.create_and_poll(assistant_id="Temperature Assistant", thread_id=thread.id)

        # Get the messages in the thread
        messages = client.beta.threads.messages.list(thread.id)

        latest_assistant_message = []
        for message in messages.data:  # Check from most recent
            if message.role == "user":
                break
            else: 
                latest_assistant_message.insert(0, message.content[0].text)

        # Display the latest assistant message
        if latest_assistant_message:
            for msg in latest_assistant_message:
                print(f"Chatbot: {msg}")
        else:
            print("No assistant response yet.")

if __name__ == "__main__":
    main()


from typing import Optional
from fastapi.responses import JSONResponse

class APIError(Exception):
    message: str
    code: int
    param: Optional[str] = None
    type: Optional[str] = 'invalid_request_error'

    def __init__(self, message: str, code: int, param: str = None, type: str = 'invalid_request_error'):
        super().__init__(message)
        self.message = message
        self.code = code
        self.param = param
        self.type = type

    def to_JSON_response(self):
        return JSONResponse({
            "error": {
                "message": self.message,
                "type": self.type,
                "param": self.param,
                "code": str(self.code)
            }
        }, status_code=self.code)


from sqlalchemy import text

def get_max_db_connections(engine):
    with engine.connect() as connection:
        result = connection.execute(text("SHOW max_connections;"))
        max_connections = result.scalar()
        return int(max_connections) if max_connections else None


from fastapi import APIRouter, HTTPException
from llamphouse.core.database.database import DatabaseManager
from ..types.message import DeleteMessageResponse, CreateMessageRequest, MessageListResponse, Attachment, MessageObject, TextContent, ImageFileContent, ModifyMessageRequest
from typing import List, Optional

router = APIRouter()

@router.post("/threads/{thread_id}/messages", response_model=MessageObject)
async def create_message(thread_id: str, request: CreateMessageRequest):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        message = db.insert_message(thread_id, request)
        return MessageObject.from_db_message(message)
     
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.get("/threads/{thread_id}/messages", response_model=MessageListResponse)
async def list_messages(thread_id: str, limit: int = 20, order: str = "desc", after: Optional[str] = None, before: Optional[str] = None):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        messages = db.list_messages_by_thread_id(
            thread_id=thread_id,
            limit=limit + 1,
            order=order,
            after=after,
            before=before
        )
        has_more = len(messages) > limit
        first_id = messages[0].id if messages else None
        last_id = messages[-1].id if messages else None

        return MessageListResponse(
            object="list",
            data=[MessageObject.from_db_message(msg) for msg in messages],
            first_id=first_id,
            last_id=last_id,
            has_more=has_more
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.get("/threads/{thread_id}/messages/{message_id}", response_model=MessageObject)
async def retrieve_message(thread_id: str, message_id: str):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        message = db.get_message_by_id(message_id)
        if not message:
            raise HTTPException(status_code=404, detail="Message not found.")
        
        return MessageObject.from_db_message(message)
    
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.post("/threads/{thread_id}/messages/{message_id}", response_model=MessageObject)
async def modify_message(thread_id: str, message_id: str, request: ModifyMessageRequest):
    try:
        db = DatabaseManager()
        message = db.update_message_metadata(thread_id, message_id, request.metadata)
        if not message:
            raise HTTPException(status_code=404, detail="Message not found.")
        
        return MessageObject.from_db_message(message)
    
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.delete("/threads/{thread_id}/messages/{message_id}", response_model=DeleteMessageResponse)
async def delete_message(thread_id: str, message_id: str):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        message = db.get_message_by_id(message_id)
        if not message:
            raise HTTPException(status_code=404, detail="Message not found.")

        deleted = db.delete_message(thread_id, message_id)
        if not deleted:
            return DeleteMessageResponse(
                id=thread_id,
                deleted=False
            )
        
        return DeleteMessageResponse(
            id=thread_id,
            deleted=True
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()



from typing import List, Optional
from abc import ABC, abstractmethod
import uuid
from .context import Context

class Assistant(ABC):
    def __init__(
        self,
        id: str,
        model: Optional[str] = "",
        name: Optional[str] = None, 
        description: Optional[str] = None, 
        temperature: Optional[float] = 0.7, 
        top_p: Optional[float] = 1.0, 
        instructions: Optional[str] = "",
        tools: Optional[List[str]] = None
    ):
        self.model = model
        self.id = id
        self.object = "assistant"
        self.name = name
        self.description = description
        self.temperature = temperature
        self.top_p = top_p
        self.instructions = instructions
        self.tools = tools or []

        # If name is provided, use it as the ID
        # If no ID and no name is provided, generate a random UUID
        if not id and name:
            self.id = name
        elif not id and not name:
            self.id = str(uuid.uuid4())
        
        if id and not name:
            self.name = id

    @abstractmethod
    async def run(self, context: Context):
        """
        Abstract method that defines the assistant's behavior.
        This should be overridden by subclasses to process the messages and threads.

        :param thread_id: The ID of the thread to run.
        :param messages: The list of messages for the assistant to process.
        :return: A response generated by the assistant.
        """
        pass


from .database.database import DatabaseManager, SessionLocal
from typing import Dict, Optional
from .types.message import Attachment, CreateMessageRequest, MessageObject
from .types.run_step import ToolCallsStepDetails
from .types.run import ToolOutput, RunObject
from .types.enum import run_step_status, run_status
import uuid
import json
import asyncio

class Context:
    def __init__(self, assistant, assistant_id: str, run_id: str, run, thread_id: str = None, queue: asyncio.Queue = None, db_session = None):
        self.assistant_id = assistant_id
        self.thread_id = thread_id
        self.run_id = run_id
        self.assistant = assistant
        self.db = DatabaseManager(db_session=db_session or SessionLocal())
        self.thread = self._get_thread_by_id(thread_id)
        self.messages = self._list_messages_by_thread_id(thread_id)
        self.run: RunObject = run
        self.__queue = queue
        
    def insert_message(self, content: str, attachment: Attachment = None, metadata: Dict[str, str] = {}, role: str = "assistant"):
        messageRequest = CreateMessageRequest(
            role=role,
            content=content,
            attachment=attachment,
            metadata=metadata
        )
        new_message = self.db.insert_message(self.thread_id, messageRequest)
        step_details = self._message_step_details(new_message.id)
        self.db.insert_run_step(run_id=self.run_id, assistant_id=self.assistant_id, thread_id=self.thread_id, step_type="message_creation", step_details=step_details, status=run_step_status.COMPLETED)

        # Update context.message
        self.messages = self._list_messages_by_thread_id(self.thread_id)
        return new_message
    
    def insert_tool_calls_step(self, step_details: ToolCallsStepDetails, output: Optional[ToolOutput] = None):
        status = run_step_status.COMPLETED if output else run_step_status.IN_PROGRESS
        run_step = self.db.insert_run_step(
            run_id=self.run_id,
            assistant_id=self.assistant_id,
            thread_id=self.thread_id,
            step_type="tool_calls",
            step_details=step_details,
            status=status
        )

        if output:
            self.db.insert_tool_output(run_step, output)
        else:
            self.db.update_run_status(self.run_id, run_status.REQUIRES_ACTION)

        return run_step
    
    def update_thread_details(self, **kwargs):
        if not self.thread:
            raise ValueError("Thread object is not initialized.")

        for key, value in kwargs.items():
            if hasattr(self.thread, key):
                setattr(self.thread, key, value)
            else:
                raise AttributeError(f"Thread object has no attribute '{key}'")
        try:
            updated_thread = self.db.update_thread(self.thread)
            return updated_thread
        except Exception as e:
            raise Exception(f"Failed to update thread in the database: {e}")

    def update_message_details(self, message_id: str, **kwargs):
        message = next((msg for msg in self.messages if msg["id"] == message_id), None)
        if not message:
            raise ValueError(f"Message with ID '{message_id}' not found in thread.")

        for key, value in kwargs.items():
            if key in message:
                message[key] = value
            else:
                raise AttributeError(f"Message object has no attribute '{key}'")

        try:
            self.db.update_message(message)
            self.messages = self._list_messages_by_thread_id(self.thread_id)
            return message
        except Exception as e:
            raise Exception(f"Failed to update message: {e}")

    def update_run_details(self, **kwargs):
        if not self.run:
            raise ValueError("Run object is not initialized.")

        for key, value in kwargs.items():
            if hasattr(self.run, key):
                setattr(self.run, key, value)
            else:
                raise AttributeError(f"Run object has no attribute '{key}'")
        try:
            updated_run = self.db.update_run(self.run)
            return updated_run
        except Exception as e:
            raise Exception(f"Failed to update run in the database: {e}")

    def _get_thread_by_id(self, thread_id):
        thread = self.db.get_thread_by_id(thread_id)
        if not thread:
            print(f"Thread with ID {thread_id} not found.")
        return thread

    def _list_messages_by_thread_id(self, thread_id):
        messages = self.db.list_messages_by_thread_id(thread_id, order="asc")
        if not messages:
            print(f"No messages found in thread {thread_id}.")
        return [MessageObject.from_db_message(msg) for msg in messages]
    
    def _get_function_from_tools(self, function_name: str):
        for tool in self.assistant.tools:
            if tool['type'] == 'function' and tool['function']['name'] == function_name:
                function_name = tool['function']['name']
                return getattr(self.assistant, function_name)
        return None

    def _message_step_details(self, message_id: str):
        return {
            "type": "message_creation",
            "message_creation": {
                "message_id": message_id
            }
        }
    
    def _function_call_step_details(self, function_name: str, args: tuple, kwargs: dict, output: str = None):
        return {
            "type": "tool_calls",
            "tool_calls": [{
                "id": str(uuid.uuid4()),
                "type": "function",
                "function": {
                    "name": function_name,
                    "arguments": {
                        "args": args,
                        "kwargs": kwargs
                    },
                    "output": output
                }
            }]
        }
    
    async def create_message_stream(self, stream):
        has_started = False
        message = None
        for chunk in stream:
            # print(f"Chunk object type:", chunk)
            if not has_started:
                has_started = True
                if chunk.object == 'chat.completion.chunk':
                    message = self.create_message('')
                    print(f"Created message:", message)
                    await self.__queue.put({"event": "thread.message.created", "data": json.dumps({
                        "id": message.id,
                        "object": "thread.message",
                        "created_at": int(message.created_at.timestamp()),
                        "thread_id": self.thread_id,
                        "role": "assistant",
                        "content": [
                            {
                            "type": "text",
                            "text": {
                                "value": "",
                                "annotations": []
                            }
                            }
                        ],
                        "assistant_id": self.assistant_id,
                        "run_id": self.run_id,
                        "attachments": [],
                        "metadata": {}
                    })})
                else:
                    print(f"Stream object type {chunk.object} not recognized.")
            else:
                if chunk.object == 'chat.completion.chunk':
                    if chunk.choices[0].delta.content == None or not chunk.choices[0].finish_reason == None:
                        await self.__queue.put({"event": "thread.message.completed", "data": json.dumps({
                            "id": message.id,
                            "object": "thread.message",
                            "created_at": int(message.created_at.timestamp()),
                            "thread_id": self.thread_id,
                            "role": "assistant",
                            "content": [
                                {
                                "type": "text",
                                "text": {
                                    "value": None,
                                    "annotations": []
                                }
                                }
                            ],
                            "assistant_id": self.assistant_id,
                            "run_id": self.run_id,
                            "attachments": [],
                            "metadata": {}
                        })})

                    # message.content.text.value += chunk.choices[0].delta.content
                        
                    await self.__queue.put({"event": "thread.message.delta", "data": json.dumps({
                        "id": message.id,
                        "object": "thread.message.delta",
                        "delta": {
                            "content": [
                                {
                                    "index": chunk.choices[0].index,
                                    "type": "text",
                                    "text": { "value": chunk.choices[0].delta.content, "annotations": [] }
                                }
                            ]
                        }
                    })})
                else:
                    print(f"Stream object type {chunk.object} not recognized.")
            # print(chunk)
            # await self.create_message(chunk)

        await self.__queue.put(None)


from llamphouse.core import LLAMPHouse
from dotenv import load_dotenv

# Import the assistant
from assistants.temperature import TemperatureAssistant

# Load the environment variables from the .env file
load_dotenv(override=True) 

def main():
    # Create an instance of the assistant
    temperature_assistant = TemperatureAssistant("Temperature Assistant")

    # Create a new LLAMPHouse instance with the assistant
    llamphouse = LLAMPHouse(assistants=[temperature_assistant], worker_type="async", api_key="secret_key")
    llamphouse.ignite(host="127.0.0.1", port=8000)

if __name__ == "__main__":
    main()


from abc import ABC, abstractmethod

class BaseWorker(ABC):
    def __init__():
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass



from fastapi import APIRouter, HTTPException, Request
from llamphouse.core.database.database import DatabaseManager
from fastapi.responses import StreamingResponse
from ..types.run import RunObject, RunCreateRequest, CreateThreadAndRunRequest, RunListResponse, ModifyRunRequest, SubmitRunToolOutputRequest
from ..types.enum import run_status, run_step_status
from ..assistant import Assistant
from typing import List, Optional
import time
import asyncio
import json

router = APIRouter()

@router.post("/threads/{thread_id}/runs", response_model=RunObject)
async def create_run(
    thread_id: str,
    request: RunCreateRequest,
    req: Request
) -> RunObject:
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        assistants = req.app.state.assistants
        assistant = get_assistant_by_id(assistants, request.assistant_id)
        # store run in db
        run = db.insert_run(thread_id, run=request, assistant=assistant)

        # Check if the task exists
        task_key = f"{request.assistant_id}:{thread_id}"
        if task_key not in req.app.state.task_queues:
            # print(f"Creating queue for task {task_key} in RUN")
            req.app.state.task_queues[task_key] = asyncio.Queue(maxsize=1)
            # raise HTTPException(status_code=404, detail="Task not found")

        # check if stream is enabled
        if request.stream:
            output_queue: asyncio.Queue = req.app.state.task_queues[task_key]

            # Streaming generator for SSE
            async def event_stream():
                while True:
                    try:
                        event = await asyncio.wait_for(output_queue.get(), timeout=10.0)  # Set timeout in seconds
                        if event is None:  # Stream completion signal
                            break
                        print(f"Event: {event['event']}")
                        # output_queue.task_done()
                        yield f"event: {event['event']}\ndata: {event['data']}\n\n"
                    except asyncio.TimeoutError:
                        print("TimeoutError: No event received within the timeout period")
                        yield f'''event: error\ndata: {json.dumps({
                            "error": "TimeoutError",
                            "message": "No event received within the timeout period"
                        })}\n\n'''
                        break

            # Return the streaming response
            return StreamingResponse(event_stream(), media_type="text/event-stream")
        
        return RunObject(
            id=run.id,
            created_at=time.mktime(run.created_at.timetuple()),
            thread_id=thread_id,
            assistant_id=assistant.id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.post("/threads/runs", response_model=RunObject)
async def create_thread_and_run(request: CreateThreadAndRunRequest, req: Request):
    try:
        db = DatabaseManager()
        thread = db.insert_thread(request.thread)

        for msg in request.thread.messages:
            if msg.role not in ["user", "assistant"]:
                raise HTTPException(status_code=400, detail="Invalid role. Must be 'user' or 'assistant'.")
            else:
                db.insert_message(thread_id=thread.id, message=msg)

        assistants = req.app.state.assistants
        assistant = get_assistant_by_id(assistants, request.assistant_id)
        # store run in db
        run = db.insert_run(thread.id, run=request, assistant=assistant)

        # Check if the task exists
        task_key = f"{request.assistant_id}:{thread.id}"
        if task_key not in req.app.state.task_queues:
            # print(f"Creating queue for task {task_key} in RUN")
            req.app.state.task_queues[task_key] = asyncio.Queue(maxsize=1)
            # raise HTTPException(status_code=404, detail="Task not found")

        # check if stream is enabled
        if request.stream:
            output_queue: asyncio.Queue = req.app.state.task_queues[task_key]

            # Streaming generator for SSE
            async def event_stream():
                while True:
                    try:
                        event = await asyncio.wait_for(output_queue.get(), timeout=10.0)  # Set timeout in seconds
                        if event is None:  # Stream completion signal
                            break
                        print(f"Event: {event['event']}")
                        # output_queue.task_done()
                        yield f"event: {event['event']}\ndata: {event['data']}\n\n"
                    except asyncio.TimeoutError:
                        print("TimeoutError: No event received within the timeout period")
                        yield f'''event: error\ndata: {json.dumps({
                            "error": "TimeoutError",
                            "message": "No event received within the timeout period"
                        })}\n\n'''
                        break

            # Return the streaming response
            return StreamingResponse(event_stream(), media_type="text/event-stream")
        
        return RunObject(
            id=run.id,
            created_at=time.mktime(run.created_at.timetuple()),
            thread=thread.id,
            assistant_id=assistant.id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()
  
def get_assistant_by_id(assistants: List[Assistant], assistant_id: str) -> Assistant:
    assistant = next((assistant for assistant in assistants if assistant.id == assistant_id), None)
    if not assistant:
        raise HTTPException(status_code=404, detail="Assistant not found")
    return assistant

@router.get("/threads/{thread_id}/runs", response_model=RunListResponse)
async def list_runs(thread_id: str, limit: int = 20, order: str = "desc", after: Optional[str] = None, before: Optional[str] = None) -> RunObject:
    try:
        db = DatabaseManager()
        runs = db.list_runs_by_thread_id(
            thread_id=thread_id,
            limit=limit + 1,
            order=order,
            after=after,
            before=before
        )
        has_more = len(runs) > limit
        first_id = runs[0].id if runs else None
        last_id = runs[-1].id if runs else None
        return  RunListResponse(
                    object="list",
                    data=[
                        RunObject(
                            id=run.id,
                            created_at=int(run.created_at.timestamp()),
                            thread_id=thread_id,
                            assistant_id=run.assistant_id,
                            status=run.status,
                            required_action=run.required_action,
                            last_error=run.last_error,
                            expires_at=run.expires_at,
                            started_at=run.started_at,
                            cancelled_at=run.cancelled_at,
                            failed_at=run.failed_at,
                            completed_at=run.completed_at,
                            incomplete_details=run.incomplete_details,
                            model=run.model,
                            instructions=run.instructions,
                            tools=run.tools,
                            metadata=run.meta,
                            usage=run.usage,
                            temperature=run.temperature,
                            top_p=run.top_p,
                            max_prompt_tokens=run.max_prompt_tokens,
                            max_completion_tokens=run.max_completion_tokens,
                            truncation_strategy=run.truncation_strategy,
                            tool_choice=run.tool_choice,
                            parallel_tool_calls=run.parallel_tool_calls,
                            response_format=run.response_format,
                        )
                        for run in runs
                    ],
                    first_id=first_id,
                    last_id=last_id,
                    has_more=has_more
                )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.get("/threads/{thread_id}/runs/{run_id}", response_model=RunObject)
async def retrieve_run(
    thread_id: str,
    run_id: str,
) -> RunObject:
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        run = db.get_run_by_id(run_id)
        if not run:
            raise HTTPException(status_code=404, detail="Run not found.")
        
        return RunObject(
            id=run.id,
            created_at=int(run.created_at.timestamp()),
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()
    
@router.post("/threads/{thread_id}/runs/{run_id}", response_model=RunObject)
async def modify_run(thread_id: str, run_id: str, request: ModifyRunRequest):
    try:
        db = DatabaseManager()
        run = db.update_run_metadata(thread_id, run_id, request.metadata)
        if not run:
            raise HTTPException(status_code=404, detail="Message not found.")
        
        return  RunObject(
            id=run.id,
            created_at=int(run.created_at.timestamp()),
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.post("/threads/{thread_id}/runs/{run_id}/submit_tool_outputs", response_model=RunObject)
async def submit_tool_outputs_to_run(thread_id: str, run_id: str, request: SubmitRunToolOutputRequest):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        run = db.get_run_by_id(run_id)
        if not run:
            raise HTTPException(status_code=404, detail="Run not found.")
        if run.status != run_status.REQUIRES_ACTION:
            raise HTTPException(status_code=400, detail="Run is not in 'requires_action' status.")

        latest_run_step = db.get_latest_run_step_by_run_id(run_id)
        if not latest_run_step:
            raise HTTPException(status_code=404, detail="No run step found for this run.")

        if not latest_run_step.step_details or "tool_calls" not in latest_run_step.step_details:
            raise HTTPException(status_code=400, detail="No tool calls found in the latest run step.")
        
        tool_calls = latest_run_step.step_details["tool_calls"]
        
        for tool_output in request.tool_outputs:
            for tool_call in tool_calls:
                if tool_call["id"] == tool_output.tool_call_id:
                    tool_call["output"] = tool_output.output
                    break

        latest_run_step.step_details = {"tool_calls": tool_calls}
        latest_run_step.status = run_step_status.COMPLETED
        db.session.commit()

        return RunObject(
            id=run.id,
            created_at=int(run.created_at.timestamp()),
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.post("/threads/{thread_id}/runs/{run_id}/cancel", response_model=RunObject)
async def cancel_run(thread_id: str, run_id: str):
    try:
        db = DatabaseManager()
        run = db.get_run_by_id(run_id)
        if not run:
            raise HTTPException(status_code=404, detail="Run not found.")
        if run.status != run_status.QUEUED:
            raise HTTPException(status_code=400, detail="Run cannot be canceled unless it is in 'queued' status.")
        
        run.status = run_status.CANCELLED
        db.session.commit()

        return RunObject(
            id=run.id,
            created_at=int(run.created_at.timestamp()),
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()


from fastapi import APIRouter, HTTPException, Request
from ..types.assistant import AssistantListResponse, AssistantObject
from typing import Optional

router = APIRouter()

@router.get("/assistants", response_model=AssistantListResponse)
async def list_assistants(
            req: Request,
            after: Optional[str] = None,
            before: Optional[str] = None,
            limit: Optional[int] = 10,
            order: Optional[str] = "asc"
        ):
    try:
        assistants = req.app.state.assistants
        if order == "desc":
            assistants = assistants[::-1]

        assistants = assistants[:limit]
        return AssistantListResponse(
            data=[AssistantObject(
                id=assistant.id,
                object=assistant.object,
                model=assistant.model,
                name=assistant.name,
                description=assistant.description,
                temperature=assistant.temperature,
                top_p=assistant.top_p,
                tools=assistant.tools
            ) for assistant in assistants],
            after=after,
            before=before
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.get("/assistants/{assistant_id}", response_model=AssistantObject)
async def retrieve_assistant(req: Request, assistant_id: str):
    try:
        assistant = next((a for a in req.app.state.assistants if a.id == assistant_id), None)
        if assistant is None:
            raise HTTPException(status_code=404, detail="Assistant not found")

        return AssistantObject(
            id=assistant.id,
            object=assistant.object,
            model=assistant.model,
            name=assistant.name,
            description=assistant.description,
            temperature=assistant.temperature,
            top_p=assistant.top_p,
            tools=assistant.tools
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")



from typing import Optional, List, Literal, Dict
from ..types.message import MessageObject, InitialMessage
from pydantic import BaseModel


class ToolResources(BaseModel):
    code_interpreter: Optional[List[str]] = None
    file_search: Optional[List[str]] = None


class ThreadObject(BaseModel):
    id: str
    created_at: int
    tool_resources: Optional[ToolResources] = None
    metadata: Optional[object] = {}
    object: Literal["thread"] = "thread"


class CreateThreadRequest(BaseModel):
    tool_resources: Optional[ToolResources] = None
    metadata: Optional[object] = {}
    messages: Optional[List[InitialMessage]] = None


class ModifyThreadRequest(BaseModel):
    tool_resources: Optional[ToolResources] = None
    metadata: Optional[object] = {}


class DeleteThreadResponse(BaseModel):
    id: str
    deleted: bool
    object: Literal["thread.deleted"] = "thread.deleted"



from typing import Optional, Union, List, Dict , Literal
from pydantic import BaseModel
from .message import Attachment, ImageFileContent, AdditionalMessage, ImageURLContent, RefusalContent, TextContent


class ToolResources(BaseModel):
    file_ids: Optional[List[str]] = None
    vector_store_ids: Optional[List[str]] = None
    

class RequiredAction(BaseModel):
    type: Optional[str]
    details: Optional[Dict]


class LastError(BaseModel):
    message: Optional[str]
    code: Optional[str]


class IncompleteDetails(BaseModel):
    reason: Optional[str]


class UsageStatistics(BaseModel):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]
    total_tokens: Optional[int]


class TruncationStrategy(BaseModel):
    type: Optional[str]
    parameters: Optional[Dict]


class ToolChoice(BaseModel):
    type: Optional[str]
    function: Optional[Dict]


class ToolCall(BaseModel):
    type: str
    function: Optional[Dict[str, str]] = None

class InitialMessage(BaseModel):
    role: str
    content: Union[str, List[Union[TextContent, ImageFileContent, ImageURLContent, RefusalContent]]]
    attachments: Optional[List[Attachment]] = None
    metadata: Optional[object] = None


class ThreadObject(BaseModel):
    messages: Optional[List[InitialMessage]]
    tool_resources: Optional[ToolResources] = None
    metadata: Optional[object] = {}


class RunObject(BaseModel):
    id: str
    created_at: int
    thread_id: str
    assistant_id: str
    status: str
    required_action: Optional[RequiredAction] = None
    last_error: Optional[LastError] = None
    expires_at: Optional[int] = None
    started_at: Optional[int] = None
    cancelled_at: Optional[int] = None
    failed_at: Optional[int] = None
    completed_at: Optional[int] = None
    incomplete_details: Optional[IncompleteDetails] = None
    model: str
    instructions: Optional[str] = None
    tools: Optional[List[Dict]] = None
    metadata: Optional[object] = {}
    object: Literal["thread.run"] = "thread.run"
    usage: Optional[UsageStatistics] = None
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    max_prompt_tokens: Optional[int] = None
    max_completion_tokens: Optional[int] = None
    truncation_strategy: Optional[TruncationStrategy] = None
    tool_choice: Optional[Union[str, ToolChoice]] = None
    parallel_tool_calls: Optional[bool] = False
    response_format: Optional[Union[str, Dict]] = "auto"


class RunCreateRequest(BaseModel):
    assistant_id: str
    model: Optional[str] = None
    instructions: Optional[Union[str, None]] = None
    additional_instructions: Optional[Union[str, None]] = None
    additional_messages: Optional[Union[List[AdditionalMessage], None]] = None
    tools: Optional[Union[List[ToolCall], None]] = None
    metadata: Optional[object] = {}
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    stream: Optional[bool] = None
    max_prompt_tokens: Optional[int] = None
    max_completion_tokens: Optional[int] = None 
    truncation_strategy: Optional[Dict[str, str]] = None  
    tool_choice: Optional[Union[str, Dict[str, str]]] = "auto"
    parallel_tool_calls: Optional[bool] = True
    response_format: Optional[Union[str, Dict[str, str]]] = "auto"


class CreateThreadAndRunRequest(BaseModel):
    assistant_id: str
    thread: Optional[ThreadObject] = None
    model: Optional[str] = None
    instructions: Optional[str] = None
    tools: Optional[List[Dict]] = None
    tool_resources: Optional[ToolResources] = None
    metadata: Optional[object] = {}
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    stream: Optional[bool] = None
    max_prompt_tokens: Optional[int] = None
    max_completion_tokens: Optional[int] = None
    truncation_strategy: Optional[TruncationStrategy] = None
    tool_choice: Optional[Union[str, ToolChoice]] = None
    parallel_tool_calls: Optional[bool] = True
    response_format: Optional[Union[str, Dict]] = "auto"

class RunListResponse(BaseModel):
    object: Literal["list"] = "list"
    data: List[RunObject]
    first_id: Optional[str] = None
    last_id: Optional[str] = None
    has_more: bool

class ModifyRunRequest(BaseModel):
    metadata: Optional[object] = {}

class ToolOutput(BaseModel):
    output: str
    tool_call_id: str

class SubmitRunToolOutputRequest(BaseModel):
    tool_outputs: List[ToolOutput]
    stream: Optional[bool] = False


from langchain_core.tools import tool
import requests

@tool
def get_coordinates_by_query(query: str) -> dict|str:
    """Get coordinates by doing a location search"""

    response = requests.get(f"https://geocoding-api.open-meteo.com/v1/search?name={query}")
    data = response.json()

    # check if a location is found
    if "results" in data and len(data["results"]) > 0:
        return data["results"][0]
    
    return f"No location found for {query}"


from langchain_core.tools import tool
import requests

@tool
def get_temperature_from_lat_lon(latitude: float, longitude: float) -> str:
    """Get the current temperature for given coordinates"""

    response = requests.get(f"https://api.open-meteo.com/v1/forecast?latitude={latitude}2&longitude={longitude}&hourly=temperature_2m")
    data = response.json()

    # check if the temperature is found
    if "hourly" in data and "time" in data["hourly"]:
        return str(data["hourly"]["temperature_2m"][0]) + "°C"

    return "Temperature not found"


from sqlalchemy import (
    Column, 
    String, 
    Text, 
    Integer, 
    ForeignKey, 
    DateTime,
    Float,
    Boolean,
    Enum
)
from sqlalchemy.sql import func
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import JSONB
import json

Base = declarative_base()


role_enum = Enum(
    'assistant', 'user', 
    name='role_enum'
)
message_status_enum = Enum(
    'in_progress', 'incomplete', 'completed', 
    name='message_status_enum'
)
run_status_enum = Enum(
    'queued', 'in_progress', 'requires_action', 'cancelling', 
    'cancelled', 'failed', 'completed', 'incomplete', 'expired', 
    name='run_status_enum'
)
step_type_enum = Enum(
    'message_creation', 'tool_calls', 
    name='run_step_type_enum'
)
run_step_status_enum = Enum(
    'in_progress', 'cancelled', 'failed', 'completed', 'expired', 
    name='run_step_status_enum'
)

class Thread(Base):
    __tablename__ = 'threads'

    id = Column(String, primary_key=True, index=True)
    name = Column(String, nullable=False)
    tool_resources = Column(JSONB)
    meta = Column(JSONB)
    updated_at = Column(DateTime, server_default=func.now(), onupdate=func.now())
    created_at = Column(DateTime, server_default=func.now())
    
    messages = relationship("Message", back_populates="thread")
    runs = relationship("Run", back_populates="thread")
    run_steps = relationship("RunStep", back_populates="thread")


class Message(Base):
    __tablename__ = 'messages'

    id = Column(String, primary_key=True, index=True)
    thread_id = Column(String, ForeignKey('threads.id'), nullable=False)
    status = Column(message_status_enum, nullable=False, server_default='in_progress')
    incomplete_details = Column(JSONB)
    role = Column(role_enum, nullable=False)
    content = Column(JSONB, nullable=False)
    assistant_id = Column(String)
    run_id = Column(String)
    attachments = Column(JSONB)
    meta = Column(JSONB)
    completed_at = Column(Integer)
    incomplete_at = Column(Integer)
    updated_at = Column(DateTime, server_default=func.now(), onupdate=func.now())
    created_at = Column(DateTime, server_default=func.now())

    thread = relationship("Thread", back_populates="messages")

    def to_json(self):
        return {
            "id": self.id,
            "object": "thread.message",
            "created_at": int(self.created_at.timestamp()) if self.created_at else None,
            "thread_id": self.thread_id,
            "role": self.role,
            "content": self.content,
            "assistant_id": self.assistant_id,
            "run_id": self.run_id,
            "attachments": self.attachments,
            "meta": self.meta,
            "completed_at": int(self.completed_at) if self.completed_at else None,
            "incomplete_at": int(self.incomplete_at) if self.incomplete_at else None,
            "status": self.status,
            "incomplete_details": self.incomplete_details
        }
    
    def __repr__(self):
        return f"<Message {self.id}>"
    
    def __str__(self):
        return json.dumps(self.to_json())

class Run(Base):
    __tablename__ = 'runs'

    id = Column(String, primary_key=True, index=True)
    status = Column(run_status_enum, nullable=False, server_default='queued')
    required_action = Column(JSONB)
    last_error = Column(JSONB)
    incomplete_details = Column(JSONB)
    model = Column(String, nullable=False)
    instructions = Column(Text, nullable=False)
    tools = Column(JSONB)
    meta = Column(JSONB)
    usage = Column(JSONB)
    temperature = Column(Float, nullable=False, server_default="1.0")
    top_p = Column(Float, nullable=False, server_default="1.0")
    max_prompt_tokens = Column(Integer)
    max_completion_tokens = Column(Integer)
    truncation_strategy = Column(JSONB)
    tool_choice = Column(JSONB)
    parallel_tool_calls = Column(Boolean, nullable=False, server_default="false")
    response_format = Column(JSONB, nullable=False, server_default='"auto"')
    thread_id = Column(String, ForeignKey('threads.id'), nullable=False)
    assistant_id = Column(String, nullable=False)
    expires_at = Column(Integer)
    started_at = Column(Integer)
    cancelled_at = Column(Integer)
    failed_at = Column(Integer)
    completed_at = Column(Integer)
    updated_at = Column(DateTime, server_default=func.now(), onupdate=func.now())
    created_at = Column(DateTime, server_default=func.now())
    
    thread = relationship("Thread", back_populates="runs")
    run_steps = relationship("RunStep", back_populates="run")

class RunStep(Base):
    __tablename__ = 'run_steps'

    id = Column(String, primary_key=True, index=True)
    object = Column(String, nullable=False, default="thread.run.step")
    created_at = Column(Integer, nullable=False)
    assistant_id = Column(String, nullable=False)
    thread_id = Column(String, ForeignKey('threads.id'), nullable=False)
    run_id = Column(String, ForeignKey('runs.id'), nullable=False)
    type = Column(step_type_enum, nullable=False)
    status = Column(run_step_status_enum, nullable=False)
    step_details = Column(JSONB)

    meta = Column(JSONB)
    usage = Column(JSONB)
    last_error = Column(JSONB)
    expired_at = Column(Integer)
    cancelled_at = Column(Integer)
    failed_at = Column(Integer)
    completed_at = Column(Integer)
    updated_at = Column(DateTime, server_default=func.now(), onupdate=func.now())
    created_at = Column(DateTime, server_default=func.now())

    thread = relationship("Thread", back_populates="run_steps")
    run = relationship("Run", back_populates="run_steps")


from .base_auth import BaseAuth

class KeyAuth(BaseAuth):

    def __init__(self, api_key: str):
        self.api_key = api_key

    def authenticate(self, api_key: str):
        return api_key == self.api_key


from fastapi import APIRouter, HTTPException
from ..types.thread import ThreadObject, CreateThreadRequest, ModifyThreadRequest, DeleteThreadResponse
from llamphouse.core.database.database import DatabaseManager
import time

router = APIRouter()

@router.post("/threads", response_model=ThreadObject)
async def create_thread(request: CreateThreadRequest):
    try:
        db = DatabaseManager()
        thread = db.insert_thread(request)
        if request.messages:
            for msg in request.messages:
                if msg.role not in ["user", "assistant"]:
                    raise HTTPException(status_code=400, detail="Invalid role. Must be 'user' or 'assistant'.")
                else:
                    db.insert_message(thread_id=thread.id, message=msg)
        
        return ThreadObject(
            id=thread.id,
            created_at=time.mktime(thread.created_at.timetuple()),
            tool_resources=thread.tool_resources,
            metadata=thread.meta
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.get("/threads/{thread_id}", response_model=ThreadObject)
async def retrieve_thread(thread_id: str):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        return ThreadObject(
            id=thread.id,
            created_at=time.mktime(thread.created_at.timetuple()),
            tool_resources=thread.tool_resources,
            metadata=thread.meta
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.post("/threads/{thread_id}", response_model=ThreadObject)
async def modify_thread(thread_id: str, request: ModifyThreadRequest):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        thread = db.update_thread_metadata(thread_id, request.metadata)
        
        return ThreadObject(
            id=thread.id,
            created_at=time.mktime(thread.created_at.timetuple()),
            tool_resources=thread.tool_resources,
            metadata=thread.meta
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()

@router.delete("/threads/{thread_id}", response_model=DeleteThreadResponse)
async def delete_thread(thread_id: str):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        db.delete_messages_by_thread_id(thread_id)
        deleted = db.delete_thread_by_id(thread_id)
        
        if not deleted:
            return DeleteThreadResponse(
                id=thread_id,
                deleted=False
            )
        
        return DeleteThreadResponse(
            id=thread_id,
            deleted=True
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()


import os
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker, Session
from .models import Thread, Message, Run, RunStep
from ..types import thread, message, run
from ..types.enum import run_status, run_step_status
from .._utils._utils import get_max_db_connections
from dotenv import load_dotenv
import uuid

load_dotenv()

DATABASE_URL = os.getenv("DATABASE_URL", "postgresql://postgres:password@localhost/llamphouse")
POOL_SIZE = int(os.getenv("POOL_SIZE", "100"))
engine = create_engine(DATABASE_URL, pool_size=int(POOL_SIZE))
SessionLocal = sessionmaker(autocommit=False, bind=engine)
MAX_POOL_SIZE = get_max_db_connections(engine)

if MAX_POOL_SIZE and POOL_SIZE > MAX_POOL_SIZE:
    raise ValueError(f"Input POOL_SIZE ({POOL_SIZE}) exceeds the database's maximum allowed ({MAX_POOL_SIZE}).")

class DatabaseManager:
    def __init__(self, db_session: Session = None):
        self.session = db_session if db_session else SessionLocal()

    def insert_thread(self, thread: thread.CreateThreadRequest):
        try:
            metadata = thread.metadata if thread.metadata else {}
            thread_id = metadata.get("thread_id", str(uuid.uuid4()))
            item = Thread(
                id=thread_id,
                name=thread_id,
                tool_resources=thread.tool_resources,
                meta=thread.metadata
            )
            self.session.add(item)
            self.session.commit()
            return item
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return None
        
    def insert_message(self, thread_id: str, message: message.CreateMessageRequest):
        try:
            metadata = message.metadata if message.metadata else {}
            message_id = metadata.get("message_id", str(uuid.uuid4()))
            item = Message(
                id=message_id,
                role=message.role,
                content=message.content,
                attachments=message.attachments,
                meta=message.metadata or {},
                thread_id=thread_id
            )
            self.session.add(item)
            self.session.commit()
            return item
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return None
        
    def insert_run(self, thread_id: str, run: run.RunCreateRequest, assistant):
        try:
            metadata = run.metadata if run.metadata else {}
            run_id = metadata.get("run_id", str(uuid.uuid4()))
            item = Run(
                id=run_id,
                thread_id=thread_id,
                assistant_id=run.assistant_id,
                model=run.model or assistant.model,
                instructions=run.instructions or assistant.instructions,
                tools=run.tools or assistant.tools,
                meta=run.metadata or {},
                temperature=run.temperature or assistant.temperature,
                top_p=run.top_p or assistant.top_p,
                max_prompt_tokens=run.max_prompt_tokens,
                max_completion_tokens=run.max_completion_tokens,
                truncation_strategy=run.truncation_strategy,
                tool_choice=run.tool_choice,
                parallel_tool_calls=run.parallel_tool_calls,
                response_format=run.response_format,
            )
            self.session.add(item)
            self.session.commit()
            return item
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return None

    def insert_run_step(self, run_id: str, assistant_id: str, thread_id: str, step_type: str, step_details: dict, status: str = run_step_status.IN_PROGRESS):
        try:
            run_step_id = str(uuid.uuid4())
            item = RunStep(
                id=run_step_id,
                object="thread.run.step",
                assistant_id=assistant_id,
                thread_id=thread_id,
                run_id=run_id,
                type=step_type,
                status=status,
                step_details=step_details
            )
            self.session.add(item)
            self.session.commit()
            self.session.refresh(item)
            return item
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return None
        
    def get_run_by_id(self, run_id: str):
        try:
            run = self.session.query(Run).filter(Run.id == run_id).first()
            return run
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def get_thread_by_id(self, thread_id: str):
        try:
            thread = self.session.query(Thread).filter(Thread.id == thread_id).first()
            return thread
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def get_message_by_id(self, message_id: str):
        try:
            message = self.session.query(Message).filter(Message.id == message_id).first()
            return message
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def get_pending_runs(self):
        try:
            pending_runs = self.session.query(Run).filter(Run.status == run_status.QUEUED).all()
            return pending_runs
        except Exception as e:
            print(f"An error occurred while fetching pending runs: {e}")
            return []
        
    def get_pending_run(self):
        try:
            pending_runs = self.session.query(Run).filter(Run.status == run_status.QUEUED).with_for_update().first()
            return pending_runs
        except Exception as e:
            print(f"An error occurred while fetching pending runs: {e}")
            return None

    def get_run_step_by_id(self, run_step_id: str) -> RunStep:
        try:
            run_step = self.session.query(RunStep).filter(RunStep.id == run_step_id).first()
            return run_step
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def get_latest_run_step_by_run_id(self, run_id: str) -> RunStep:
        try:
            run_step = (
                self.session.query(RunStep)
                .filter(RunStep.run_id == run_id)
                .order_by(RunStep.created_at.desc())
                .first()
            )
            return run_step
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def insert_tool_output(self, run_step: RunStep, tool_output: run.ToolOutput):
        try:
            tool_calls = run_step.step_details["tool_calls"]
            for tool_call in tool_calls:
                if tool_call["id"] == tool_output.tool_call_id:
                    tool_call["output"] = tool_output.output
                    break

            run_step.step_details = {"tool_calls": tool_calls}
            run_step.status = run_step_status.COMPLETED
            self.session.commit()
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def list_messages_by_thread_id(
            self,
            thread_id: str, 
            limit: int = 20, 
            order: str = "desc", 
            after: str = None, 
            before: str = None
        ) -> list[Message]:
        try:
            query = self.session.query(Message).filter(Message.thread_id == thread_id)
            if order == "asc":
                query = query.order_by(Message.created_at.asc())
            else:
                query = query.order_by(Message.created_at.desc())
            if after:
                query = query.filter(Message.id > after)
            if before:
                query = query.filter(Message.id < before)
            return query.limit(limit).all()
        except Exception as e:
            print(f"An error occurred: {e}")
            return []
        
    def list_runs_by_thread_id(
            self,
            thread_id: str, 
            limit: int = 20, 
            order: str = "desc", 
            after: str = None, 
            before: str = None
        ) -> list[Message]:
        try:
            query = self.session.query(Run).filter(Run.thread_id == thread_id)
            if order == "asc":
                query = query.order_by(Run.created_at.asc())
            else:
                query = query.order_by(Run.created_at.desc())
            if after:
                query = query.filter(Run.id > after)
            if before:
                query = query.filter(Run.id < before)
            return query.limit(limit).all()
        except Exception as e:
            print(f"An error occurred: {e}")
            return []

    def list_run_steps(self, 
            thread_id: str, 
            run_id: str,
            limit: int = 20, 
            order: str = "desc", 
            after: str = None, 
            before: str = None
        ) -> list[RunStep]:
        try:
            query = self.session.query(RunStep).filter(RunStep.run_id == run_id, RunStep.thread_id == thread_id)
            if order == "asc":
                query = query.order_by(RunStep.created_at.asc())
            else:
                query = query.order_by(RunStep.created_at.desc())
            if after:
                query = query.filter(RunStep.id > after)
            if before:
                query = query.filter(RunStep.id < before)
            return query.limit(limit).all()
        except Exception as e:
            print(f"An error occurred while fetching run steps: {e}")
            return []

    def update_thread_metadata(self, thread_id: str, metadata: dict):
        try:
            thread = self.session.query(Thread).filter(Thread.id == thread_id).first()
            if thread:
                thread.meta = metadata
                self.session.commit()
                return thread
            return None
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating thread metadata: {e}")
            return None

    def update_message_metadata(self, thread_id: str, message_id: str, metadata: dict):
        try:
            message = self.session.query(Message).filter(Message.thread_id == thread_id, Message.id == message_id).first()
            
            if message:
                message.meta = metadata or {}
                self.session.commit()
                return message
            else:
                return None
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return None

    def update_thread(self, thread: Thread):
        try:
            self.session.merge(thread)
            self.session.commit()
            return thread
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating the thread: {e}")
            return None 

    def update_message(self, message: Message):
        try:
            self.session.merge(message)
            self.session.commit()
            return message
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating the message: {e}")
            return None

    def update_run(self, run: Run):
        try:
            self.session.merge(run)
            self.session.commit()
            return run
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating the run: {e}")
            return None
        
    def update_run_metadata(self, thread_id: str, run_id: str, metadata: dict):
        try:
            run = self.session.query(Run).filter(Run.thread_id == thread_id, Run.id == run_id).first()
            if run:
                run.meta = metadata
                self.session.commit()
                return run
            return None
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating thread metadata: {e}")
            return None
        
    def update_run_status(self, run_id: str, status: str, error: dict = None):
        try:
            run = self.session.query(Run).filter(Run.id == run_id).first()
            if run:
                run.status = status
                run.last_error = error
                self.session.commit()
                return run
            return None
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating the run: {e}")
            return None
        
    def update_run_step_status(self, run_step_id: str, status: str, output = None, error: str = None):
        try:
            run_step = self.session.query(RunStep).filter(RunStep.id == run_step_id).first()
            if run_step:
                run_step.status = status
                run_step.last_error = error
                if output:
                    run_step.step_details["tool_calls"][0]["function"]["output"] = output
                self.session.commit()
                return run_step
            return None
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred while updating the run step: {e}")
            return None

    def delete_thread_by_id(self, thread_id: str):
        try:
            thread = self.session.query(Thread).filter(Thread.id == thread_id).first()
            if thread:
                self.session.delete(thread)
                self.session.commit()
                return True
            return False
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return False

    def delete_messages_by_thread_id(self, thread_id: str):
        try:
            self.session.query(Message).filter(Message.thread_id == thread_id).delete()
            self.session.commit()
            return True
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return False

    def delete_message_by_id(self, message_id: str):
        try:
            message = self.session.query(Message).filter(Message.id == message_id).first()
            if message:
                self.session.delete(message)
                self.session.commit()
                return True
            return False
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return False

    def delete_message(self, thread_id: str, message_id: str):
        try:
            message = self.session.query(Message).filter(
                Message.thread_id == thread_id, Message.id == message_id).first()
            
            if message:
                self.session.delete(message)
                self.session.commit()
                return True
            else:
                return False
        except Exception as e:
            self.session.rollback()
            print(f"An error occurred: {e}")
            return False


from .base_worker import BaseWorker
from concurrent.futures import ThreadPoolExecutor
from ..database.database import sessionmaker, engine
from ..types.enum import run_status
from ..database.models import Run
from ..context import Context
import queue
import time
import threading

class ThreadWorker(BaseWorker):
    def __init__(self, time_out=30, thread_count=1):
        self.thread_count = thread_count
        self.time_out = time_out
        
        self.SessionLocal = sessionmaker(autocommit=False, bind=engine)
        self.threads = []
        self.running = True

    def task_execute(self):
        while self.running:
            try:
                session = self.SessionLocal()
                task = (
                    session.query(Run)
                    .filter(Run.status == run_status.QUEUED)
                    .with_for_update(skip_locked=True)
                    .first()
                )

                if task:
                    task.status = run_status.IN_PROGRESS
                    session.commit()

                    assistant = next(
                        (assistant for assistant in self.assistants if assistant.id == task.assistant_id),
                        None
                    )
                    if not assistant:
                        task.status = run_status.FAILED
                        task.last_error = {
                            "code": "server_error",
                            "message": "Assistant not found"
                        }
                        session.commit()
                        continue
                    
                    task_key = f"{task.assistant_id}:{task.thread_id}"
                    output_queue = queue.Queue()
                    self.fastapi_state.task_queues[task_key] = output_queue
                    context = Context(assistant=assistant, assistant_id=task.assistant_id, thread_id=task.thread_id, run_id=task.id, run=task, queue=output_queue, db_session=session)
                    with ThreadPoolExecutor(max_workers=1) as executor:
                        future = executor.submit(assistant.run, context)
                        try:
                            future.result(timeout=self.time_out)
                            task.status = run_status.COMPLETED
                            session.commit()
                        except TimeoutError:
                            task.status = run_status.EXPIRED
                            task.last_error = {
                                "code": "timeout_error",
                                "message": f"Task execution exceeded the {self.time_out}-second timeout."
                            }
                            session.commit()
                        except Exception as e:
                            task.status = run_status.FAILED
                            task.last_error = {
                                "code": "server_error",
                                "message": str(e)
                            }
                            session.commit()
                else:
                    time.sleep(1)
            except Exception as e:
                session.rollback()
                print(f"ThreadWorker error: {e}")
            finally:
                session.close()

    def start(self, **kwargs):
        self.assistants = kwargs.get("assistants", [])
        self.fastapi_state = kwargs.get("fastapi_state", {})

        for i in range(self.thread_count):
            thread = threading.Thread(target=self.task_execute)
            thread.start()
            self.threads.append(thread)

            print(f"ThreadWorker thread {i} started")

    def stop(self):
        self.running = False
        for thread in self.threads:
            thread.join()


from typing import Optional, Union, List, Literal
from .tool_call import ToolCall
from pydantic import BaseModel

class MessageCreation(BaseModel):
    message_id: str

class ToolCallsStepDetails(BaseModel):
    tool_calls: List[ToolCall]
    type: Literal["tool_calls"]

class MessageCreationStepDetails(BaseModel):
    message_creation: MessageCreation
    type: Literal["message_creation"]


class StepDetails(BaseModel):
    Union[MessageCreationStepDetails, ToolCallsStepDetails]

class LastError(BaseModel):
    code: Literal["server_error", "rate_limit_exceeded"]
    message: str


class Usage(BaseModel):
    completion_tokens: int
    prompt_tokens: int
    total_tokens: int

class RunStepObject(BaseModel):
    id: str
    assistant_id: str
    cancelled_at: Optional[int] = None
    completed_at: Optional[int] = None
    created_at: int
    expired_at: Optional[int] = None
    failed_at: Optional[int] = None
    last_error: Optional[LastError] = None
    metadata: Optional[object] = None
    object: Literal["thread.run.step"] = "thread.run.step"
    run_id: str
    status: Literal["in_progress", "cancelled", "failed", "completed", "expired"] = "completed"
    step_details: StepDetails
    thread_id: str
    type: Literal["message_creation", "tool_calls"]
    usage: Optional[Usage] = None
    
class RunStepListResponse(BaseModel):
    object: Literal["list"] = "list"
    data: List[RunStepObject]
    first_id: Optional[str] = None
    last_id: Optional[str] = None
    has_more: bool

class RunStepInclude(BaseModel):
    Literal["step_details.tool_calls[*].file_search.results[*].content"]

class RunStepListRequest(BaseModel):
    limit: Optional[int] = 20
    order: Optional[str] = "desc"
    after: Optional[str] = None
    before: Optional[str] = None
    include: Optional[List[RunStepInclude]]



from llamphouse.core import Assistant
from llamphouse.core.context import Context
from langchain_core.messages import HumanMessage, AIMessage, ToolMessage
from langchain_openai.chat_models import ChatOpenAI
from langchain_core.utils.function_calling import convert_to_openai_tool
from langgraph.graph import StateGraph, START, END
from typing import Optional, Annotated
from typing_extensions import TypedDict

# import the tools
from tools.get_coordinates_by_query import get_coordinates_by_query
from tools.get_temperature_from_lat_lon import get_temperature_from_lat_lon

def add_messages(messages: list[HumanMessage | AIMessage], data: list[HumanMessage | AIMessage]) -> list[HumanMessage | AIMessage]:
    return data # return the new list of messages

# Define the state of the graph
class State(TypedDict):
    messages: Annotated[list[HumanMessage | AIMessage], add_messages]
    latitude: float
    longitude: float
    user_location_name: Optional[float] = None

# LangGraph node for executing tools
class BasicToolNode:
    """A node that runs the tools requested in the last AIMessage."""

    def __init__(self, tools: list) -> None:
        self.tools_by_name = {tool.name: tool for tool in tools}

    def __call__(self, inputs: dict):
        if messages := inputs.get("messages", []):
            message = messages[-1]
        else:
            raise ValueError("No message found in input")

        for tool_call in message.tool_calls:

            if tool_call["name"] == get_coordinates_by_query.name:
                tool_result = self.tools_by_name[tool_call["name"]].invoke(
                    tool_call["args"]
                )

                inputs["user_location_name"] = tool_call["args"].get("query", None)

                if "latitude" in tool_result and "longitude" in tool_result:
                    inputs["latitude"] = tool_result["latitude"]
                    inputs["longitude"] = tool_result["longitude"]

            elif tool_call["name"] == get_temperature_from_lat_lon.name:
                tool_result = self.tools_by_name[tool_call["name"]].invoke(
                    tool_call["args"]
                )
                
            inputs["messages"].append(
                ToolMessage(
                    content=tool_result,
                    name=tool_call["name"],
                    tool_call_id=tool_call["id"],
                )
            )

        return inputs

def route_tools(state):
    """
    Use in the conditional_edge to route to the ToolNode if the last message
    has tool calls. Otherwise, route to the end.
    """
    if isinstance(state, list):
        ai_message = state[-1]
    elif messages := state.get("messages", []):
        ai_message = messages[-1]
    else:
        raise ValueError(f"No messages found in input state to tool_edge: {state}")
    if hasattr(ai_message, "tool_calls") and len(ai_message.tool_calls) > 0:
        return "tools"
    return END

# Define the llamphouse assistant
class TemperatureAssistant(Assistant):
    def __init__(self, id: str):
        super().__init__(model="gpt-4o", id=id, instructions="I can help you with temperature queries.")

    def run(self, context: Context, *args, **kwargs):

        chat_messages = []
        for msg in reversed(context.messages):
            if msg.role == "user":
                chat_messages.append(HumanMessage(content=msg.content))
            elif msg.role == "assistant":
                chat_messages.append(AIMessage(content=msg.content))
            elif msg.role == 'system':
                continue
            else:
                raise Exception(f"Unknown role encountered in chat history: {msg.role}.")

        # Define the available tools
        tools = [get_coordinates_by_query, get_temperature_from_lat_lon]

        llm = ChatOpenAI(temperature=0)

        def chatbot(state: State):
            llm_with_tools = llm.bind_tools(tools=[convert_to_openai_tool(tool) for tool in tools])
            response = llm_with_tools.invoke(state["messages"])
            return {
                "messages": state["messages"] + [response]
            }

        state = State(messages=chat_messages)

        # Define the graph structure 
        graph_builder = StateGraph(State)

        # Add the tools node to the graph
        tool_node = BasicToolNode(tools=tools)
        graph_builder.add_node("tools", tool_node)

        # Add the chatbot node to the graph
        graph_builder.add_node("chatbot", chatbot)

        # Add the start and end nodes to the graph
        graph_builder.add_edge(START, "chatbot")
        graph_builder.add_edge("chatbot", END)

        # The `tools_condition` function returns "tools" if the chatbot asks to use a tool, and "END" if
        # it is fine directly responding. This conditional routing defines the main agent loop.
        graph_builder.add_conditional_edges(
            "chatbot",
            route_tools,
            {"tools": "tools", END: END},
        )
        # Any time a tool is called, we return to the chatbot to decide the next step
        graph_builder.add_edge("tools", "chatbot")

        graph = graph_builder.compile()

        response = graph.invoke(state)

        last_assistant_message = [msg for msg in response['messages'] if type(msg) == AIMessage][-1]

        context.create_message(last_assistant_message.content)


import sys
from typing import List, Optional
import uvicorn
from fastapi import FastAPI
from .routes import all_routes
from .assistant import Assistant
from .workers.base_worker import BaseWorker
from .workers.async_worker import AsyncWorker
from .middlewares.catch_exceptions_middleware import CatchExceptionsMiddleware
from .middlewares.auth_middleware import AuthMiddleware
from .auth.base_auth import BaseAuth
import asyncio

class LLAMPHouse:
    def __init__(self, assistants: List[Assistant] = [], authenticator: Optional[BaseAuth] = None, worker: Optional[BaseWorker] = AsyncWorker()):
        self.assistants = assistants
        self.worker = worker
        self.authenticator = authenticator
        self.fastapi = FastAPI(title="LLAMPHouse API Server")
        self.fastapi.state.assistants = assistants
        self.fastapi.state.task_queues = {}

        # Add middlewares
        self.fastapi.add_middleware(CatchExceptionsMiddleware)
        if self.authenticator:
            self.fastapi.add_middleware(AuthMiddleware, auth=self.authenticator)

        self._register_routes()

    def __print_ignite(self, host, port):
        ascii_art = """
                  __,--'
       .-.  __,--'
      |  o| 
     [IIIII]`--.__
      |===|       `--.__
      |===|
      |===|
      |===|
______[===]______
"""
        print(ascii_art)
        print("We have light!")
        print(f"LLAMPHOUSE server running on http://{host}:{port}")
        sys.stdout.flush()

    def ignite(self, host="0.0.0.0", port=80, reload=False):
        
        @self.fastapi.on_event("startup")
        async def startup_event():
            loop = asyncio.get_event_loop()
            self.worker.start(assistants=self.assistants, fastapi_state=self.fastapi.state, loop=loop)

        @self.fastapi.on_event("shutdown")
        async def on_shutdown():
            print("Server shutting down...")
            if self.worker:
                print("Stopping worker...")
                self.worker.stop() 
        self.__print_ignite(host, port)
        uvicorn.run(self.fastapi, host=host, port=port, reload=reload)

    def _register_routes(self):       
        for router in all_routes:
            self.fastapi.include_router(router)



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse
from .._exceptions import APIError

class CatchExceptionsMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        try:
            response = await call_next(request)

        except APIError as e:
            response = e.to_JSON_response()

        except Exception as e:
            response = JSONResponse({
                "error": {
                    "message": str(e),
                    "type": "unhandled_internal_server_error",
                    "param": None,
                    "code": None
                }
            }, status_code=500)

        return response


from typing import Optional, Union, List, Literal
from pydantic import BaseModel

class CodeInterpreterOutputLogs(BaseModel):
    logs: str
    type: Literal["logs"]

class CodeInterpreterOutputImageImage(BaseModel):
    file_id: str

class CodeInterpreterOutputImage(BaseModel):
    image: CodeInterpreterOutputImageImage
    type: Literal["image"]

class CodeInterpreterOutput(BaseModel):
    Union[CodeInterpreterOutputLogs, CodeInterpreterOutputImage]

class CodeInterpreter(BaseModel):
    input: str
    outputs: List[CodeInterpreterOutput]

class CodeInterpreterToolCall(BaseModel):
    id: str
    code_interpreter: CodeInterpreter
    type: Literal["code_interpreter"]

class FileSearchRankingOptions(BaseModel):
    ranker: Literal["default_2024_08_21"]
    score_threshold: float

class FileSearchResultContent(BaseModel):
    text: Optional[str] = None
    type: Optional[Literal["text"]] = None

class FileSearchResult(BaseModel):
    file_id: str
    file_name: str
    score: float
    content: Optional[List[FileSearchResultContent]] = None

class FileSearch(BaseModel):
    ranking_options: Optional[FileSearchRankingOptions] = None
    results: Optional[List[FileSearchResult]] = None

class FileSearchToolCall(BaseModel):
    id: str
    file_search: FileSearch
    type: Literal["file_search"]

class Function(BaseModel):
    arguments: str
    name: str
    output: Optional[str] = None

class FunctionToolCall(BaseModel):
    id: str
    function: Function
    type: Literal["function"]

class ToolCall(BaseModel):
    Union[CodeInterpreterToolCall, FileSearchToolCall, FunctionToolCall]



import asyncio
from ..database.database import engine
from sqlalchemy.orm import sessionmaker
from ..types.enum import run_status
from .base_worker import BaseWorker
from ..assistant import Assistant
from ..database.models import Run
from ..context import Context
from typing import List, Optional

class AsyncWorker(BaseWorker):
    def __init__(self,time_out=30):
        """
        Initialize the AsyncWorker.

        Args:
            session_factory: A factory function to create database sessions.
            assistants: List of assistant objects for processing runs.
            fastapi_state: Shared state object from the FastAPI application.
            timeout: Timeout for processing each run (in seconds).
            sleep_interval: Time to sleep between checking the queue (in seconds).
        """
        self.time_out = time_out

        # self.assistants: List[Assistant] = kwargs.get("assistants", [])
        # self.fastapi_state = kwargs.get("fastapi_state", {})
        # self.loop = kwargs.get("loop", None)
        # if not self.loop:
        #     raise ValueError("loop is required")

        self.task = None
        self.SessionLocal = sessionmaker(autocommit=False, bind=engine)
        self.running = True

        print("AsyncWorker initialized")

    async def process_run_queue(self):
        """
        Continuously process the run queue, fetching and handling pending runs.
        """
        while self.running:
            try:
                session = self.SessionLocal()
                run = (
                    session.query(Run)
                    .filter(Run.status == run_status.QUEUED)
                    .with_for_update(skip_locked=True)
                    .first()
                )

                if run:
                    run.status = run_status.IN_PROGRESS
                    session.commit()
                    
                    assistant = next((assistant for assistant in self.assistants if assistant.id == run.assistant_id), None)
                    if not assistant:
                        run.status = run_status.FAILED
                        run.last_error = {
                            "code": "server_error",
                            "message": "Assistant not found"
                        }
                        session.commit()
                        continue

                    task_key = f"{run.assistant_id}:{run.thread_id}"

                    if task_key not in self.fastapi_state.task_queues:
                        # print(f"Creating queue for task {task_key}")
                        self.fastapi_state.task_queues[task_key] = asyncio.Queue(maxsize=1)

                    output_queue = self.fastapi_state.task_queues[task_key]

                    context = Context(assistant=assistant, assistant_id=run.assistant_id, run_id=run.id, run=run, thread_id=run.thread_id, queue=output_queue, db_session=session)

                    try:
                        await asyncio.wait_for(
                            asyncio.to_thread(assistant.run, context),
                            timeout=self.time_out
                        )
                        run.status = run_status.COMPLETED
                        session.commit()

                    except asyncio.TimeoutError:
                        print(f"Run {run.id} timed out.")
                        run.status = run_status.EXPIRED
                        run.last_error = {
                            "code": "server_error",
                            "message": "Run timeout"
                        }
                        session.commit()


                    except Exception as e:
                        print(f"Error executing run {run.id}: {e}")
                        run.status = run_status.FAILED
                        run.last_error = {
                            "code": "server_error",
                            "message": str(e)
                        }
                        session.commit()

                    print(f"Run {run.id} completed.")

            except Exception as e:
                print(f"Error processing run queue: {e}")

            finally:
                session.close()
                # Sleep for a short period to avoid tight loops if there are no pending runs
                await asyncio.sleep(2)


    def start(self, **kwargs):
        """
        Start the async worker to process the run queue.
        """
        self.assistants = kwargs.get("assistants", [])
        self.fastapi_state = kwargs.get("fastapi_state", {})
        self.loop = kwargs.get("loop", None)
        if not self.loop:
            raise ValueError("loop is required")
        
        self.task = self.loop.create_task(self.process_run_queue())

    def stop(self):
        print("Stopping async worker...")
        self.running = False
        if self.task:
            self.task.cancel()



from typing import Optional, Union, Dict, List, Literal
from pydantic import BaseModel

class AssistantObject(BaseModel):
    id: str
    object: Literal["assistant"] = "assistant"
    model: str
    name: Optional[str] = None
    description: Optional[str] = None
    temperature: Optional[float] = 0.7 
    top_p: Optional[float] = 1.0
    instruction: Optional[str] = None
    tools: Optional[List[str]] = None

class AssistantListResponse(BaseModel):
    data: List[AssistantObject]
    after: Optional[str]
    before: Optional[str]

class AssistantListRequest(BaseModel):
    after: Optional[str] = None
    before: Optional[str] = None
    limit: Optional[int] = 20
    order: Optional[str] = "desc"

class AssistantCreateRequest(BaseModel):
    model: str
    description: Optional[str] = None
    instructions: Optional[str] = None
    metadata: Optional[object] = {}
    name: Optional[str] = None
    response_format: Optional[Union[str, Dict]] = "auto"
    temperature: Optional[float] = 1.0
    tool_resources: Optional[dict] = None
    tools: Optional[List[str]] = []
    top_p: Optional[float] = 1.0
    extra_headers: Optional[dict] = None
    extra_query: Optional[dict] = None
    extra_body: Optional[dict] = None
    timeout: Optional[float] = None

class AssistantCreateResponse(BaseModel):
    id: str
    model: str
    name: str
    description: Optional[str]
    instructions: Optional[str]
    tools: List[str]
    file_ids: List[str]
    metadata: Optional[object] = {}
    object: str
    temperature: float
    top_p: float
    response_format: Union[str, Dict]

class ModifyAssistantRequest(BaseModel):
    model: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    instructions: Optional[str] = None
    tools: Optional[List[Dict[str, str]]] = []
    file_ids: Optional[List[str]] = []
    metadata: Optional[object] = {}
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    response_format: Optional[Union[str, Dict[str, str]]] = "auto"


from .assistant import router as assistant_router
from .threads import router as threads_router
from .message import router as message_router
from .run import router as run_router
from .run_step import router as run_step_router

all_routes = [assistant_router, threads_router, message_router, run_router, run_step_router]



from llamphouse.core import LLAMPHouse, Assistant
from dotenv import load_dotenv
from llamphouse.core.context import Context
from openai import OpenAI
from llamphouse.core.auth.base_auth import BaseAuth

load_dotenv(override=True)

open_client = OpenAI()

class CustomAuth(BaseAuth):
    def authenticate(self, api_key: str):
        if api_key == "secret_key":
            return True
        return False

# Create a custom assistant
class CustomAssistant(Assistant):

    def run(self, context: Context):
        # transform the assistant messages to chat messages
        messages = [{"role": message.role, "content": message.content[0].text} for message in context.messages]
        
        # send the messages to the OpenAI API
        result = open_client.chat.completions.create(
            messages=messages,
            model="gpt-4o-mini"
        )

        # add the assistant messages to the thread
        context.insert_message(role="assistant", content=result.choices[0].message.content)

        # no need to return anything, the run will stop here

def main():
    # Create an instance of the custom assistant
    my_assistant = CustomAssistant("my-assistant")

    # Create a new LLAMPHouse instance
    llamphouse = LLAMPHouse(assistants=[my_assistant], authenticator=CustomAuth())
    
    # Start the LLAMPHouse server
    llamphouse.ignite(host="127.0.0.1", port=8000)

if __name__ == "__main__":
    main()



from pydantic import BaseModel
from typing import Optional, Union, Dict, List

class CompletionCreateRequest(BaseModel):
    model: str
    prompt: str
    best_of: Optional[int] = 1
    echo: Optional[bool] = False 
    max_tokens: Optional[int] = 16
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    logprobs: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    logit_bias: Optional[Dict[int, float]] = None
    stream: Optional[bool] = False 
    user: Optional[str] = None
    seed: Optional[int] = None
    suffix: Optional[str] = None
    stream_options: Optional[dict] = None

class CompletionTokensDetails(BaseModel):
    accepted_prediction_tokens: int
    audio_tokens: int
    reasoning_tokens: int
    rejected_prediction_tokens: int

class PromptTokensDetails(BaseModel):
    audio_tokens: int
    cached_tokens: int

class Usage(BaseModel):
    completion_tokens: int
    prompt_tokens: int
    total_tokens: int
    completion_tokens_details: CompletionTokensDetails
    prompt_tokens_details: PromptTokensDetails

class Logprobs(BaseModel):
    text_offset: list[int] = []
    token_logprobs: list[float] = []
    tokens: list[str] = []
    top_logprobs: list[dict] = []

class Choice(BaseModel):
    text: str
    index: int
    finish_reason: str
    logprobs: Logprobs = None

class CompletionCreateResponse(BaseModel):
    id: str
    object: str
    created: int
    model: str
    system_fingerprint: str
    choices: Choice
    usage: Usage


QUEUED = "queued"
IN_PROGRESS = "in_progress"
REQUIRES_ACTION = "requires_action"
CANCELLING = "cancelling"
CANCELLED = "cancelled"
FAILED = "failed"
COMPLETED = "completed"
INCOMPLETE = "incomplete"
EXPIRED = "expired"


from setuptools import setup, find_namespace_packages

# Read the contents of requirements.txt
with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

def get_version():
    with open("llamphouse/__init__.py", "r") as f:
        for line in f:
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

setup(
    name="llamphouse",
    version=get_version(),
    author="llamp.ai",
    author_email="info@llamp.ai",
    description="LLAMPHouse OpenAI Assistant Server",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=['llamphouse.core', 'llamphouse.core.*']),
    python_requires='>=3.10',
    install_requires=install_requires,
    package_data={},
)


from fastapi import APIRouter, HTTPException
from llamphouse.core.database.database import DatabaseManager
from ..types.run_step import RunStepObject, RunStepListResponse
from typing import Optional

router = APIRouter()


@router.get("/threads/{thread_id}/runs/{run_id}/steps", response_model=RunStepListResponse)
async def list_run_steps(thread_id: str, run_id: str, limit: int = 20, order: str = "desc", after: Optional[str] = None, before: Optional[str] = None):
    try:
        db = DatabaseManager()
        run_steps = db.list_run_steps(
            thread_id=thread_id,
            run_id=run_id,
            limit=limit + 1,
            order=order,
            after=after,
            before=before
        )
        print(len(run_steps))
        has_more = len(run_steps) > limit
        first_id = run_steps[0].id if run_steps else None
        last_id = run_steps[-1].id if run_steps else None
        return  RunStepListResponse(
                    object="list",
                    data=[
                        RunStepObject(
                            id=run_step.id,
                            assistant_id=run_step.assistant_id,
                            cancelled_at=run_step.cancelled_at,
                            completed_at=run_step.completed_at,
                            created_at=int(run_step.created_at.timestamp()),
                            expires_at=run_step.expired_at,
                            failed_at=run_step.failed_at,
                            last_error=run_step.last_error,
                            metadata=run_step.meta,
                            run_id=run_step.run_id,
                            status=run_step.status,
                            step_details=run_step.step_details,
                            thread_id=run_step.thread_id,
                            type=run_step.type,
                            usage=run_step.usage,
                        )
                        for run_step in run_steps
                    ],
                    first_id=first_id,
                    last_id=last_id,
                    has_more=has_more
                )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()



@router.get("/threads/{thread_id}/runs/{run_id}/steps/{step_id}", response_model=RunStepObject)
async def retrieve_run_step(thread_id: str, run_id: str, step_id: str):
    try:
        db = DatabaseManager()
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        run = db.get_run_by_id(run_id)
        if not run:
            raise HTTPException(status_code=404, detail="Run not found.")
        
        run_step = db.get_run_step_by_id(step_id)
        if not run_step:
            raise HTTPException(status_code=404, detail="Run step not found.")
        return  RunStepObject(
                    id=run_step.id,
                    assistant_id=run_step.assistant_id,
                    cancelled_at=run_step.cancelled_at,
                    completed_at=run_step.completed_at,
                    created_at=int(run_step.created_at.timestamp()),
                    expires_at=run_step.expired_at,
                    failed_at=run_step.failed_at,
                    last_error=run_step.last_error,
                    metadata=run_step.meta,
                    run_id=run_step.run_id,
                    status=run_step.status,
                    step_details=run_step.step_details,
                    thread_id=run_step.thread_id,
                    type=run_step.type,
                    usage=run_step.usage,
                )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    finally:
        db.session.close()



from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from .._exceptions import APIError
from ..auth.base_auth import BaseAuth

class AuthMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, auth: BaseAuth):
        super().__init__(app)
        self.auth = auth

    async def dispatch(self, request: Request, call_next):
        auth_header = request.headers.get("Authorization")
        if not auth_header or not auth_header.startswith("Bearer "):
            return APIError(code=401, message="Missing or invalid API key.").to_JSON_response()

        api_key = auth_header.split("Bearer ")[1]
        if not self.auth.authenticate(api_key):
            return APIError(code=403, message="Invalid API key.").to_JSON_response()

        return await call_next(request)



from typing import Optional, List, Dict, Literal, Union
from pydantic import BaseModel

class Attachment(BaseModel):
    file_id: str
    tool: Optional[str] = None

class IncompleteDetails(BaseModel):
    reason: str
    details: Optional[Dict[str, Union[str, int]]] = None

class TextContent(BaseModel):
    type: Literal["text"] = "text"
    text: str

class ImageFileContent(BaseModel):
    type: Literal["image_file"] = "image_file"
    image_file: str

class ImageURLContent(BaseModel):
    type: Literal["image_url"] = "image_url"
    image_url: str

class RefusalContent(BaseModel):
    type: Literal["refusal"] = "refusal"
    refusal_text: str

class InitialMessage(BaseModel):
    role: str
    content: Union[str, List[Union[TextContent, ImageFileContent, ImageURLContent, RefusalContent]]]
    attachments: Optional[List[Attachment]] = None
    metadata: Optional[object] = None

class AdditionalMessage(BaseModel):
    role: str
    content: Union[str, List[Union[TextContent, ImageFileContent, ImageURLContent, RefusalContent]]]
    attachments: Optional[List[Attachment]] = None
    metadata: Optional[object] = None

class MessageObject(BaseModel):
    id: str
    created_at: int
    thread_id: str
    status: Literal["in_progress", "incomplete", "completed"] = "completed"
    incomplete_details: Optional[IncompleteDetails] = None
    completed_at: Optional[int] = None
    incomplete_at: Optional[int] = None
    role: str
    content: List[Union[TextContent, ImageFileContent, ImageURLContent, RefusalContent]]
    assistant_id: Optional[str] = None
    run_id: Optional[str] = None
    attachments: Optional[List[Attachment]] = None
    metadata: Optional[object] = None
    object: Literal["thread.message"] = "thread.message"

    @staticmethod
    def from_db_message(message) -> "MessageObject":
        return MessageObject(
            id=message.id,
            role=message.role,
            content=[
                TextContent(text=message.content) if isinstance(message.content, str) else ImageFileContent(image_file=message.content)
                for message.content in [message.content]
            ],
            metadata=message.meta,
            status=message.status,
            incomplete_details=message.incomplete_details,
            completed_at=message.completed_at,
            incomplete_at=message.incomplete_at,
            assistant_id=message.assistant_id,
            run_id=message.run_id,
            attachments=[
                Attachment(file_id=attachment['file_id'], tool=attachment.get('tool')) for attachment in (message.attachments or [])
            ],
            created_at=int(message.created_at.timestamp()),
            thread_id=message.thread_id
        )

class CreateMessageRequest(BaseModel):
    role: str
    content: str
    attachments: Optional[Attachment] = None
    metadata: Optional[object] = None

class MessagesListRequest(BaseModel):
    limit: Optional[int] = 20
    order: Optional[str] = "desc"
    after: Optional[str] = None
    before: Optional[str] = None
    run_id: Optional[str] = None

class MessageListResponse(BaseModel):
    object: Literal["list"] = "list"
    data: List[MessageObject]
    first_id: Optional[str] = None
    last_id: Optional[str] = None
    has_more: bool

class ModifyMessageRequest(BaseModel):
    metadata: Optional[object] = {}

class DeleteMessageResponse(BaseModel):
    id: str
    deleted: bool
    object: Literal["thread.message.deleted"] = "thread.message.deleted"


from openai import OpenAI

def main():
    # Connect to the LLAMPHouse server
    client = OpenAI(api_key="secret_key", base_url="http://127.0.0.1:8000")
    
    # Create new thread
    empty_thread = client.beta.threads.create()
    print(f"Created thread: {empty_thread.id}")

    # Add message to thread
    client.beta.threads.messages.create(empty_thread.id, role="user", content="This is my first message")
    client.beta.threads.messages.create(empty_thread.id, role="user", content="How are you")

    # Create new run in specific thread & assistant
    new_run =client.beta.threads.runs.create_and_poll(assistant_id="my-assistant", thread_id=empty_thread.id, temperature=1)
    print(f"Run finished: {new_run}")

if __name__ == "__main__":
    main()
